# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities Resource API."""

from flask import g
from flask_resources import resource_requestctx, response_handler, route
from invenio_records_resources.resources.files.resource import request_stream
from invenio_records_resources.resources.records.resource import \
    RecordResource, request_data, request_headers, request_search_args, \
    request_view_args
from invenio_records_resources.resources.records.utils import es_preference


class CommunityResource(RecordResource):
    """Communities resource."""

    def create_url_rules(self):
        """Create the URL rules for the record resource."""

        def p(prefix, route):
            """Prefix a route with the URL prefix."""
            return f"{prefix}{route}"

        routes = self.config.routes
        return [
            route(
                "GET",
                p(routes["communities-prefix"], routes["list"]),
                self.search
            ),
            route(
                "POST",
                p(routes["communities-prefix"], routes["list"]),
                self.create
            ),
            route(
                "GET",
                p(routes["communities-prefix"], routes["item"]),
                self.read
            ),
            route(
                "PUT",
                p(routes["communities-prefix"], routes["item"]),
                self.update
            ),
            route(
                "DELETE",
                p(routes["communities-prefix"], routes["item"]),
                self.delete
            ),
            route(
                "GET",
                p(routes["user-prefix"], routes["list"]),
                self.search_user_communities
            ),
            route(
                "POST",
                p(routes["communities-prefix"], routes["item"]) + '/rename',
                self.rename
            ),
            route(
                "GET",
                p(routes["communities-prefix"], routes["item"]) + '/logo',
                self.read_logo
            ),
            route(
                "PUT",
                p(routes["communities-prefix"], routes["item"]) + '/logo',
                self.update_logo
            ),
            route(
                "DELETE",
                p(routes["communities-prefix"], routes["item"]) + '/logo',
                self.delete_logo
            ),
            route(
                "GET",
                p(routes["communities-prefix"], routes["featured-prefix"]),
                self.featured_communities_search
            ),
            route(
                "GET",
                p(routes["communities-prefix"], routes["item"]) + routes["featured-prefix"],
                self.featured_list
            ),
            route(
                "POST",
                p(routes["communities-prefix"], routes["item"]) + routes["featured-prefix"],
                self.featured_create
            ),
            route(
                "PUT",
                p(routes["communities-prefix"], routes["item"]) + p(routes["featured-prefix"], routes["featured-id"]),
                self.featured_update
            ),
            route(
                "DELETE",
                p(routes["communities-prefix"], routes["item"]) + p(routes["featured-prefix"], routes["featured-id"]),
                self.featured_delete
            ),
        ]

    @request_search_args
    @response_handler(many=True)
    def search_user_communities(self):
        """Perform a search over the user's communities.

        GET /user/communities
        """
        hits = self.service.search_user_communities(
            identity=g.identity,
            params=resource_requestctx.args,
            es_preference=es_preference()
        )
        return hits.to_dict(), 200

    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def rename(self):
        """Rename a community."""
        item = self.service.rename(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data,
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return item.to_dict(), 200

    @request_view_args
    def read_logo(self):
        """Read logo's content."""
        item = self.service.read_logo(
            g.identity,
            resource_requestctx.view_args["pid_value"],
        )
        return item.send_file(), 200

    @request_view_args
    @request_stream
    @response_handler()
    def update_logo(self):
        """Upload logo content."""
        item = self.service.update_logo(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data["request_stream"],
            content_length=resource_requestctx.data["request_content_length"],
        )
        return item.to_dict(), 200

    @request_view_args
    def delete_logo(self):
        """Delete logo."""
        self.service.delete_logo(
            g.identity,
            resource_requestctx.view_args["pid_value"],
        )
        return "", 204

    @request_search_args
    @response_handler(many=True)
    def featured_communities_search(self):
        hits = self.service.featured_search(
            identiy=g.identity,
            params=resource_requestctx.args,
            es_preference=es_preference(),
        )
        return hits.to_dict(), 200

    @request_headers
    @request_view_args
    @response_handler()
    def featured_list(self):
        """List featured entries for a community."""
        items = self.service.featured_list(
            g.identity,
            resource_requestctx.view_args["pid_value"],
        )
        return items.to_dict(), 200


    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def featured_create(self):
        """Create a featured community entry."""
        item = self.service.featured_create(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data,
        )
        return item.to_dict(), 201

    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def featured_update(self):
        """Update a featured community entry."""
        item = self.service.featured_update(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data,
            featured_id=resource_requestctx.view_args["featured_id"],
        )
        return item.to_dict(), 200

    @request_view_args
    def featured_delete(self):
        """Delete a featured community entry."""
        self.service.featured_delete(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            featured_id=resource_requestctx.view_args["featured_id"],
        )
        return "", 204



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community PID provider."""

from invenio_pidstore.errors import PIDAlreadyExists, PIDDoesNotExistError
from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.base import BaseProvider


class CommunitiesIdProvider(BaseProvider):
    """Community identifier provider.

    This is the recommended community id provider.

    It uses the value of the 'id' present in our data to generate the
    identifier.
    """

    pid_type = 'comid'
    """Type of persistent identifier."""

    pid_provider = None
    """Provider name."""

    object_type = 'rec'
    """Type of object."""

    default_status = PIDStatus.REGISTERED
    """Community IDs with an object are by default registered.

    Default: :attr:`invenio_pidstore.models.PIDStatus.REGISTERED`
    """

    @classmethod
    def create(cls, record, **kwargs):
        """Create a new commuinity identifier.

        For more information about parameters,
        see :meth:`invenio_pidstore.providers.base.BaseProvider.create`.

        :param record: The community record.
        :param kwargs: dict to hold generated pid_value and status. See
            :meth:`invenio_pidstore.providers.base.BaseProvider.create` extra
            parameters.
        :returns: A :class:`CommunitiesIdProvider` instance.
        """
        kwargs['pid_value'] = record['id']
        kwargs['status'] = cls.default_status
        kwargs['object_type'] = cls.object_type
        kwargs['object_uuid'] = record.model.id
        return super(CommunitiesIdProvider, cls).create(**kwargs)

    @classmethod
    def update(cls, pid, new_value):
        """`Update the value of the Community identifier`.

        :param pid: Persistent Identifier type.
        :param new_value: The new string value.

        :returns: A :class:`CommunitiesIdProvider` instance.
        """
        try:
            existing_pid = cls.get(new_value).pid
        except PIDDoesNotExistError:
            pass
        else:
            raise PIDAlreadyExists(
                existing_pid.pid_type,
                existing_pid.pid_value
            )
        pid.pid_value = new_value
        return cls(pid)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""API views."""


def create_communities_api_blueprint(app):
    """Create communities api blueprint."""
    ext = app.extensions["invenio-communities"]
    # control blueprint endpoints registration
    return ext.communities_resource.as_blueprint()


def create_members_api_bp_from_app(app):
    """Create members api blueprint."""
    ext = app.extensions["invenio-communities"]
    # control blueprint endpoints registration
    return ext.members_resource.as_blueprint()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import itertools
from copy import deepcopy

import pytest
from flask_principal import AnonymousIdentity
from invenio_access.permissions import any_user as any_user_need
from invenio_accounts.models import Role
from invenio_app.factory import create_api
from invenio_requests.proxies import current_requests_service

from invenio_communities.communities.records.api import Community
from invenio_communities.proxies import current_communities

pytest_plugins = ("celery.contrib.pytest", )


#
# Application
#
@pytest.fixture(scope='module')
def app_config(app_config):
    """Override pytest-invenio app_config fixture."""
    app_config['RECORDS_REFRESOLVER_CLS'] = \
        "invenio_records.resolver.InvenioRefResolver"
    app_config['RECORDS_REFRESOLVER_STORE'] = \
        "invenio_jsonschemas.proxies.current_refresolver_store"
    # Variable not used. We set it to silent warnings
    app_config['JSONSCHEMAS_HOST'] = 'not-used'

    return app_config


@pytest.fixture(scope="module")
def create_app(instance_path):
    """Application factory fixture."""
    return create_api


@pytest.fixture()
def headers():
    """Default headers for making requests."""
    return {
        'content-type': 'application/json',
        'accept': 'application/json',
    }


#
# Services
#
@pytest.fixture(scope="module")
def community_service(app):
    """Community service."""
    return current_communities.service


@pytest.fixture(scope="module")
def member_service(community_service):
    """Members subservice."""
    return community_service.members


@pytest.fixture(scope="module")
def requests_service(app):
    """Requests service."""
    return current_requests_service


#
# Users and groups
#
@pytest.fixture(scope="module")
def anon_identity():
    """A new user."""
    identity = AnonymousIdentity()
    identity.provides.add(any_user_need)
    return identity


@pytest.fixture(scope="module")
def users(UserFixture, app, database):
    users = {}
    for r in ['owner', 'manager', 'curator', 'reader']:
        u = UserFixture(
            email=f'{r}@{r}.org',
            password=r,
        )
        u.create(app, database)
        users[r] = u
    return users


@pytest.fixture(scope="module")
def group(database):
    r = Role(name='it-dep')
    database.session.add(r)
    database.session.commit()
    return r


@pytest.fixture(scope="module")
def owner(users):
    """Community owner user."""
    return users['owner']


@pytest.fixture(scope="module")
def any_user(UserFixture, app, database):
    """A user without privileges or memberships."""
    u = UserFixture(
        email=f'anyuser@anyuser.org',
        password='anyuser',
    )
    u.create(app, database)
    u.identity # compute identity
    return u


#
# Communities
#
@pytest.fixture(scope="module")
def minimal_community():
    """Minimal community metadata."""
    return  {
        "access": {
            "visibility": "public",
            "record_policy": "open",
        },
        "id": "public",
        "metadata": {
            "title": "My Community",
        }
    }

@pytest.fixture(scope="module")
def full_community():
    """Full community data as dict coming from the external world."""
    return  {
        "access": {
            "visibility": "public",
            "record_policy": "open",
        },
        "id": "my_community_id",
        "metadata": {
            "title": "My Community",
            "description": "This is an example Community.",
            "type": "event",
            "curation_policy": "This is the kind of records we accept.",
            "page": "Information for my community.",
            "website": "https://inveniosoftware.org/",
            "organizations": [{
                    "name": "CERN",
            }]
        }
    }


@pytest.fixture(scope="module")
def community(community_service, owner, minimal_community, location):
    """A community."""
    c =  community_service.create(owner.identity, minimal_community)
    owner.refresh()
    return c


@pytest.fixture(scope="module")
def restricted_community(community_service, owner, minimal_community, location):
    """A restricted community."""
    data = deepcopy(minimal_community)
    data['access']['visibility'] = 'restricted'
    data['id'] = 'restricted'
    c =  community_service.create(owner.identity, data)
    owner.refresh()
    return c


@pytest.fixture(scope="function")
def fake_communities(community_service, owner, minimal_community, location, db):
    """Multiple community created and posted to test search functionality."""
    community_types = ['organization', 'event', 'topic', 'project']

    N = 4
    for (type_,ind) in itertools.product(community_types, list(range(N))):
        minimal_community['id'] = f'comm_{type_}_{ind}'
        minimal_community['metadata']['type'] = type_
        c =  community_service.create(owner.identity, minimal_community)
    Community.index.refresh()

    # Return ids of first and last created communities
    return 'comm_organization_0', 'comm_project_3', N, N*len(community_types)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test community members resources."""

import pytest


#
# Fixtures
#
@pytest.fixture()
def community_id(community):
    return community._record.id


@pytest.fixture(scope="function")
def group_data(group):
    return {
        "members": [{"type": "group", "id": group.name}],
        "role": "reader",
    }


@pytest.fixture(scope="function")
def new_user_data(new_user):
    return {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
    }


#
# Add
#
def test_add(client, headers, community_id, owner, group_data, db):
    """Test add REST API."""
    client = owner.login(client)
    r = client.post(
        f'/communities/{community_id}/members',
        headers=headers,
        json=group_data
    )
    assert r.status_code == 204


def test_add_denied(client, headers, community_id, group_data, new_user):
    """Test add REST API."""
    client = new_user.login(client)
    r = client.post(
        f'/communities/{community_id}/members',
        headers=headers,
        json=group_data
    )
    assert r.status_code == 403


def test_add_bad_data(client, headers, community_id, owner):
    """Test add REST API."""
    client = owner.login(client)
    r = client.post(
        f'/communities/{community_id}/members',
        headers=headers,
        json={"members": []}
    )
    assert r.status_code == 400


def test_add_invalid_member(client, headers, community_id, owner, group_data):
    """Test add REST API."""
    client = owner.login(client)
    group_data['members'][0]['id'] = 'invalid'
    r = client.post(
        f'/communities/{community_id}/members',
        headers=headers,
        json=group_data
    )
    assert r.status_code == 400


def test_add_duplicate(client, headers, community_id, owner, group_data, db):
    """Test add REST API."""
    client = owner.login(client)
    r = client.post(
        f'/communities/{community_id}/members',
        headers=headers,
        json=group_data
    )
    assert r.status_code == 204
    r = client.post(
        f'/communities/{community_id}/members',
        headers=headers,
        json=group_data
    )
    assert r.status_code == 400


#
# Invite
#
def test_invite(client, headers, community_id, owner, new_user_data, db):
    """Test invite REST API."""
    client = owner.login(client)
    r = client.post(
        f'/communities/{community_id}/invitations',
        headers=headers,
        json=new_user_data,
    )
    assert r.status_code == 204


def test_invite_deny(
    client, headers, community_id, new_user, new_user_data, db):
    """Test invite REST API."""
    client = new_user.login(client)
    r = client.post(
        f'/communities/{community_id}/invitations',
        headers=headers,
        json=new_user_data
    )
    assert r.status_code == 403


#
# Update
#
def test_update(
    client, headers, community_id, owner, public_reader):
    """Test update of members."""
    client = owner.login(client)
    data = {
        "members": [{"type": "user", "id": str(public_reader.id)}],
        "visible": False,
        "role": "curator",
    }
    r = client.put(
        f'/communities/{community_id}/members',
        headers=headers,
        json=data,
    )
    assert r.status_code == 204


def test_update_invite(
    client, headers, community_id, owner, new_user_data, db):
    """Test update of members."""
    client = owner.login(client)
    r = client.post(
        f'/communities/{community_id}/invitations',
        headers=headers,
        json=new_user_data,
    )
    assert r.status_code == 204

    # Update the invite
    new_user_data["role"] = "curator"
    r = client.put(
        f'/communities/{community_id}/invitations',
        headers=headers,
        json=new_user_data,
    )
    assert r.status_code == 204


#
# Delete
#
def test_delete(
    client, headers, community_id, owner, public_reader):
    """Test delete of members."""
    client = public_reader.login(client)
    data = {
        "members": [{"type": "user", "id": str(public_reader.id)}],
    }
    r = client.delete(
        f'/communities/{community_id}/members',
        headers=headers,
        json=data,
    )
    assert r.status_code == 204


#
# Search and serialization
#
def test_search(
    client, headers, community_id, owner, public_reader, db, clean_index):
    """Search."""
    client = owner.login(client)
    r = client.get(
        f'/communities/{community_id}/members',
        headers=headers,
    )
    assert r.status_code == 200
    data = r.json
    assert data['hits']['total'] == 2
    assert 'role' in data['aggregations']
    assert 'visibility' in data['aggregations']

    hit = data['hits']['hits'][0]
    assert 'role' in hit
    assert 'visible' in hit
    assert 'created' in hit
    assert 'updated' in hit
    assert 'revision_id' in hit
    assert 'is_current_user' in hit
    assert 'permissions' in hit
    assert hit['permissions']['can_leave'] is False
    assert hit['permissions']['can_delete'] is True
    assert hit['permissions']['can_update_role'] is True
    assert hit['permissions']['can_update_visible'] is True
    assert hit['visible'] is True

    hit = data['hits']['hits'][1]
    assert hit['is_current_user'] is True
    assert hit['permissions']['can_leave'] is True
    assert hit['permissions']['can_delete'] is False
    assert hit['permissions']['can_update_role'] is False
    assert hit['permissions']['can_update_visible'] is True
    assert hit['visible'] is False

    # Pagination params should be passed correctly as well.
    # see https://github.com/inveniosoftware/invenio-communities/issues/495
    r1 = client.get(
        f'/communities/{community_id}/members',
        headers=headers,
        query_string={"page": 1, "size": 1},
    ).json
    assert r1['hits']['total'] == 2
    assert len(r1['hits']['hits']) == 1

    r2 = client.get(
        f'/communities/{community_id}/members',
        headers=headers,
        query_string={"page": 2, "size": 1},
    ).json
    assert r2['hits']['total'] == 2
    assert len(r2['hits']['hits']) == 1
    assert r1['hits']['hits'][0]["id"] != r2['hits']['hits'][0]


def test_search_public(
    client, headers, community_id, new_user, public_reader, clean_index):
    """Search public members."""
    client = new_user.login(client)
    r = client.get(
        f'/communities/{community_id}/members/public',
        headers=headers,
    )
    assert r.status_code == 200
    data = r.json
    assert data['hits']['total'] == 1
    hit = data['hits']['hits'][0]
    # Public view has no facets (because that would leak information on
    # roles/visible)
    assert 'aggregations' not in data
    # A member in the public view should not leak below attributes:
    assert 'role' not in hit
    assert 'visible' not in hit
    assert 'created' not in hit
    assert 'updated' not in hit
    assert 'revision_id' not in hit
    assert 'is_current_user' not in hit
    assert 'permissions' not in hit
    # A member do have:
    assert 'member' in hit
    assert 'id' in hit['member']
    assert 'type' in hit['member']
    assert 'name' in hit['member']
    assert 'description' in hit['member']


def test_search_invitation(
    client, headers, community_id, owner, invite_user, db, clean_index):
    """Search invitations."""
    client = owner.login(client)
    r = client.get(
        f'/communities/{community_id}/invitations',
        headers=headers,
    )
    assert r.status_code == 200
    data = r.json
    assert data['hits']['total'] == 1
    assert 'role' in data['aggregations']
    assert 'status' in data['aggregations']
    hit = data['hits']['hits'][0]
    assert 'role' in hit
    assert 'visible' in hit
    assert 'created' in hit
    assert 'updated' in hit
    assert 'revision_id' in hit
    assert 'request' in hit
    assert 'id' in hit['request']
    assert 'status' in hit['request']
    assert 'expires_at' in hit['request']
    assert hit['request']['expires_at'] is not None
    assert 'permissions' in hit
    assert hit['permissions']['can_update_role'] is True
    assert hit['permissions']['can_cancel'] is True


# TODO: member serialization/links
# TODO: request serialization/links
# TODO: community member can see info
# TODO: community non-member can't see info
# TODO: facet by role, facet by visibility, define sorts.
# TODO: same user can be invited to two different communities
# TODO: same user/group can be added to two different communities



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Members."""

from .records import Member, MemberModel
from .resources import MemberResource, MemberResourceConfig
from .services import MemberService, MemberServiceConfig


__all__ = (
    'Member',
    'MemberModel',
    'MemberService',
    'MemberServiceConfig',
    'MemberResource',
    'MemberResourceConfig',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Members Service Config."""

from flask_babelex import lazy_gettext as _
from invenio_records_resources.services import RecordServiceConfig, \
    SearchOptions, pagination_links

from ...communities.records.api import Community
from ...permissions import CommunityPermissionPolicy
from ..records import Member
from . import facets
from .schemas import MemberEntitySchema


class PublicSearchOptions(SearchOptions):
    # TODO: should restrict fields that's being searched on.
    # query_parser_cls = QueryParser
    sort_default = 'bestmatch'
    # TODO: sort options should be by username and not expose e.g. creation
    # date
    sort_default_no_query = 'newest'
    sort_options = {
        "bestmatch": dict(
            title=_('Best match'),
            fields=['_score'],  # ES defaults to desc on `_score` field
        ),
        "newest": dict(
            title=_('Newest'),
            fields=['-created'],
        ),
    }


class InvitationsSearchOptions(SearchOptions):
    # TODO: should restrict fields that's being searched on.
    # query_parser_cls = QueryParser
    sort_default = 'bestmatch'
    # TODO: sort options should be by username and not expose e.g. creation
    # date
    sort_default_no_query = 'newest'
    sort_options = {
        "bestmatch": dict(
            title=_('Best match'),
            fields=['_score'],  # ES defaults to desc on `_score` field
        ),
        "newest": dict(
            title=_('Newest'),
            fields=['-created'],
        ),
        "oldest": dict(
            title=_('Oldest'),
            fields=['created'],
        ),
    }

    facets = {
        'role': facets.role,
        'status': facets.status,
    }


class MemberSearchOptions(PublicSearchOptions):
    """Search options."""
    sort_default = 'bestmatch'
    sort_default_no_query = 'newest'
    sort_options = {
        "bestmatch": dict(
            title=_('Best match'),
            fields=['_score'],  # ES defaults to desc on `_score` field
        ),
        # TODO add user name
        "newest": dict(
            title=_('Newest'),
            fields=['-created'],
        ),
        "oldest": dict(
            title=_('Oldest'),
            fields=['created'],
        ),
    }

    facets = {
        'role': facets.role,
        'visibility': facets.visibility,
    }


class MemberServiceConfig(RecordServiceConfig):
    """Member Service Config."""

    community_cls = Community
    record_cls = Member
    schema = MemberEntitySchema

    permission_policy_cls = CommunityPermissionPolicy

    search = MemberSearchOptions
    search_public = PublicSearchOptions
    search_invitations = InvitationsSearchOptions

    # No links
    links_item = {}

    # ResultList configurations
    links_search = pagination_links(
        "{+api}/communities/{community_id}/members{?args*}"
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Member views."""

from .config import MemberResourceConfig
from .resource import MemberResource

__all__ = (
    'MemberResource',
    'MemberResourceConfig',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test community member service."""

import pytest
from invenio_access.permissions import system_identity
from invenio_records_resources.services.errors import PermissionDeniedError
from marshmallow import ValidationError

from invenio_communities.members.errors import AlreadyMemberError, \
    InvalidMemberError
from invenio_communities.members.records.api import ArchivedInvitation, Member


#
# Add members
#
@pytest.mark.parametrize("actor,role", [
    ('owner', 'owner'),
    ('owner', 'manager'),
    ('owner', 'curator'),
    ('owner', 'reader'),
    ('manager', 'manager'),
    ('manager', 'curator'),
    ('manager', 'reader'),
])
def test_add_allowed(member_service, community, members, group, actor, role, db):
    """Test that the given roles CAN add a group member."""
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": role,
    }
    member_service.add(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize("actor,role", [
    ('manager', 'owner'),
    ('curator', 'owner'),
    ('curator', 'manager'),
    ('curator', 'curator'),
    ('curator', 'reader'),
    ('reader', 'owner'),
    ('reader', 'manager'),
    ('reader', 'curator'),
    ('reader', 'reader'),
])
def test_add_denied(
        member_service, community, members, group, actor, role, db):
    """Test that the given roles CANNOT add a group member."""
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": role,
    }
    assert pytest.raises(
        PermissionDeniedError,
        member_service.add,
        members[actor].identity,
        community._record.id,
        data
    )


def test_add_duplicate(member_service, community, owner, group, db):
    """Test that we cannot add duplicate members."""
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "reader",
    }
    member_service.add(owner.identity, community._record.id, data)
    data["role"] = "curator"
    # Duplicate detected (even if role changed)
    assert pytest.raises(
        AlreadyMemberError,
        member_service.add, owner.identity, community._record.id, data
    )


def test_add_invalid_member_type(
        member_service, community, owner, new_user, db):
    """Only system identity can add a denied member type."""
    data = {
        # "type: user" is allowed to add only by system
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
    }
    assert pytest.raises(
        PermissionDeniedError,
        member_service.add, owner.identity, community._record.id, data
    )
    member_service.add(system_identity, community._record.id, data)


def test_add_visible_property(
        member_service, community, owner, new_user, db):
    """Only system identity can set visible property."""
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
        # visible allowed only to be set by system.
        "visible": True,
    }
    assert pytest.raises(
        ValidationError,
        member_service.add, owner.identity, community._record.id, data
    )
    member_service.add(system_identity, community._record.id, data)


def test_add_invalid_data(member_service, community, owner, group, db):
    """Test various forms of invalid data."""
    # Invalid group id
    data = {"members": [{"type": "group", "id": "invalid"}], "role": "reader"}
    assert pytest.raises(
        InvalidMemberError,
        member_service.add, owner.identity, community._record.id, data
    )
    # Missing member id
    data = {"members": [{"type": "group"}], "role": "reader"}
    assert pytest.raises(
        ValidationError,
        member_service.add, owner.identity, community._record.id, data
    )
    # Missing member type
    data = {"members": [{"id": "groupname"}], "role": "reader"}
    assert pytest.raises(
        ValidationError,
        member_service.add, owner.identity, community._record.id, data
    )
    # Invalid member type
    data = {"members": [{"type": "invalid", "id": "1"}], "role": "reader"}
    assert pytest.raises(
        ValidationError,
        member_service.add, owner.identity, community._record.id, data
    )
    # No members
    data = {"members": [], "role": "reader"}
    assert pytest.raises(
        Exception,
        member_service.add, owner.identity, community._record.id, data
    )
    # Invalid role
    data = {"members": [{"type": "group", "id": group.name}], "role": "inval"}
    assert pytest.raises(
        ValidationError,
        member_service.add, owner.identity, community._record.id, data
    )
    # Cannot add email
    data = {
        "members": [{"type": "email", "id": "somebody@somewhere.org"}],
        "role": "reader",
    }
    assert pytest.raises(
        ValidationError,
        # using system_identity because owner gets a permission denied before
        # validation kicks in.
        member_service.add, system_identity, community._record.id, data
    )


#
# Invite members
#
def test_invite(
        member_service, community, owner, new_user, db):
    """Invite a user."""
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
    }
    member_service.invite(owner.identity, community._record.id, data)
    # ensure that the invited user request has been indexed
    res = member_service.search_invitations(
        owner.identity, community._record.id).to_dict()
    assert res['hits']['total'] == 1
    # Cannot invite twice.
    pytest.raises(
        AlreadyMemberError,
        member_service.invite, owner.identity, community._record.id, data
    )
    # Cannot add either if invited
    pytest.raises(
        AlreadyMemberError,
        member_service.add, system_identity, community._record.id, data
    )


def test_invite_group_denied(
        member_service, community, owner, group, db):
    """Invite a group."""
    # Groups cannot be invited (groups cannot receive invitation request)
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "reader",
    }
    # - owner identity
    assert pytest.raises(
        PermissionDeniedError,
        member_service.invite, owner.identity, community._record.id, data
    )
    # - system identity
    assert pytest.raises(
        InvalidMemberError,
        member_service.invite, system_identity, community._record.id, data
    )
    # Email invitations not yet implemented
    data["members"][0]["type"] = "email"
    assert pytest.raises(
        ValidationError,
        member_service.invite, owner.identity, community._record.id, data
    )


def test_invite_already_member(
        member_service, community, owner, new_user, db):
    """Invite a user."""
    res = member_service.search(
        owner.identity, community._record.id).to_dict()
    current = res['hits']['total']

    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
    }
    member_service.add(system_identity, community._record.id, data)
    res = member_service.search(
        owner.identity, community._record.id).to_dict()
    assert res['hits']['total'] == current + 1
    # Cannot invite if already a member.
    pytest.raises(
        AlreadyMemberError,
        member_service.invite, owner.identity, community._record.id, data
    )


def test_invite_with_message(
        member_service, community, owner, new_user, db):
    """Invite a user."""
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
        "message": "Welcome to the club!"
    }
    member_service.invite(owner.identity, community._record.id, data)
    # Invalid message
    data["message"] = 1
    pytest.raises(
        ValidationError,
        member_service.invite, owner.identity, community._record.id, data
    )


def test_invite_view_request(requests_service, invite_user, db, clean_index):
    """A request should have been created."""
    res = requests_service.search(
        invite_user.identity,
        receiver={'user': invite_user.id},
        type='community-invitation',
    ).to_dict()
    assert res['hits']['total'] == 1


#
# Search and read members
#
def test_search_members(
        member_service, community, owner, public_reader, anon_identity,
        clean_index):
    """Members can see all members, anyone can only see public."""
    # Members can see all other members.
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()['hits']['total'] == 2

    # Anyone else can only see public members.
    res = member_service.search_public(anon_identity, community._record.id)
    assert res.to_dict()['hits']['total'] == 1


def test_search_members_restricted(
        member_service, restricted_community, owner, anon_identity,
        clean_index):
    """Restricted communities can only be searched by members."""
    c = restricted_community

    # Members can see all other members.
    res = member_service.search(owner.identity, c._record.id)
    assert res.to_dict()['hits']['total'] == 1

    # Anyone get permission denied.
    pytest.raises(
        PermissionDeniedError,
        member_service.search_public, anon_identity, c._record.id
    )


#
# Search invitations
#
def test_search_invitations(
        member_service, requests_service, community, owner, invite_user,
        invite_request_id, db, clean_index):
    """Search invitations should include archived invitations."""
    # Decline the invitation and reinvite
    requests_service.execute_action(
        invite_user.identity, invite_request_id, 'decline').to_dict()
    data = {
        "members": [{"type": "user", "id": str(invite_user.id)}],
        "role": "reader",
    }
    member_service.invite(owner.identity, community._record.id, data)

    # See invitations in list
    res = member_service.search_invitations(
        owner.identity, community._record.id)
    assert res.to_dict()['hits']['total'] == 2


#
# Accept/decline/cancel invite
#
def test_invite_accept_flow(
        member_service, requests_service, community, owner, invite_user,
        invite_request_id, db, clean_index):
    """Invite a user."""
    # Accept request
    request = requests_service.execute_action(
        invite_user.identity, invite_request_id, 'accept').to_dict()

    # See new member in list
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()['hits']['total'] == 2

    # See invitation in archived list
    ArchivedInvitation.index.refresh()
    res = member_service.search_invitations(
        owner.identity, community._record.id)
    assert res.to_dict()['hits']['total'] == 1


def test_invite_decline_flow(
        member_service, requests_service, community, owner, invite_user,
        invite_request_id, db, clean_index):
    """Invite a user."""
    # Decline request
    request = requests_service.execute_action(
        invite_user.identity, invite_request_id, 'decline').to_dict()

    # Only owner in list
    Member.index.refresh()
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()['hits']['total'] == 1

    # It's possible to reinvite
    data = {
        "members": [{"type": "user", "id": str(invite_user.id)}],
        "role": "reader",
    }
    member_service.invite(owner.identity, community._record.id, data)


def test_invite_cancel_flow(
        member_service, requests_service, community, owner, invite_request_id,
        db, clean_index):
    """Invite a user."""
    # Cancel request
    request = requests_service.execute_action(
        owner.identity, invite_request_id, 'cancel').to_dict()

    # Only owner in list
    Member.index.refresh()
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()['hits']['total'] == 1


def test_invite_actions_permissions(
        requests_service, owner, any_user, members, invite_user,
        invite_request_id, db):
    """Actions should be protected."""
    manager = members['manager']
    curator = members['curator']
    reader = members['reader']

    # Only invited user can accept/decline
    for user in [owner, any_user, manager, curator, reader]:
        pytest.raises(
            PermissionDeniedError,
            requests_service.execute_action,
            user.identity, invite_request_id, 'accept'
        )
        pytest.raises(
            PermissionDeniedError,
            requests_service.execute_action,
            user.identity, invite_request_id, 'decline'
        )

    # Only community owners/managers can cancel
    for user in [any_user, curator, reader, invite_user]:
        pytest.raises(
            PermissionDeniedError,
            requests_service.execute_action,
            user.identity, invite_request_id, 'cancel'
        )


#
# Leave community
#
@pytest.mark.parametrize("role", ['manager', 'curator', 'reader'])
def test_leave_allowed(member_service, community, members, role):
    """Managers, curators and readers can leave."""
    user = members[role]
    data = {'members': [{"type": "user", 'id': str(user.id)}]}
    assert member_service.delete(
        user.identity,
        community._record.id,
        data
    )


def test_leave_single_owner_denied(member_service, community, owner):
    """A single owner cannot leave"""
    data = {'members': [{"type": "user", 'id': str(owner.id)}]}
    pytest.raises(
        ValidationError,
        member_service.delete,
        owner.identity,
        community._record.id,
        data
    )


def test_leave_denied(member_service, community, any_user, invite_user):
    """No permission for others"""
    data = {'members': [{"type": "user", 'id': str(any_user.id)}]}
    pytest.raises(
        PermissionDeniedError,
        member_service.delete,
        any_user.identity,
        community._record.id,
        data
    )
    data = {'members': [{"type": "user", 'id': str(invite_user.id)}]}
    pytest.raises(
        PermissionDeniedError,
        member_service.delete,
        invite_user.identity,
        community._record.id,
        data
    )


def test_leave_owner_allowed(member_service, community, owner, group, db):
    """If multiple owners exists, an owner can leave"""
    # Add an owner
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "owner",
    }
    member_service.add(owner.identity, community._record.id, data)

    # Leave
    data = {'members': [{"type": "user", 'id': str(owner.id)}]}
    assert member_service.delete(
        owner.identity, community._record.id, data)


#
# Delete members
#
# TODO: there should be one user owner (group owner not enough).
@pytest.mark.parametrize("actor,role", [
    ('manager', 'owner'),
    ('curator', 'owner'),
    ('curator', 'manager'),
    ('curator', 'curator'),
    ('curator', 'reader'),
    ('reader', 'owner'),
    ('reader', 'manager'),
    ('reader', 'curator'),
    ('reader', 'reader'),
])
def test_delete_denied(
        member_service, community, members, actor, new_user, role, db):
    """Test that the given roles CANNOT be removed by the actor."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Delete the user
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
    }
    assert pytest.raises(
        PermissionDeniedError,
        member_service.delete,
        members[actor].identity,
        community._record.id,
        data
    )


@pytest.mark.parametrize("actor,role", [
    ('owner', 'owner'),
    ('owner', 'manager'),
    ('owner', 'curator'),
    ('owner', 'reader'),
    ('manager', 'manager'),
    ('manager', 'curator'),
    ('manager', 'reader'),
])
def test_delete_allowed(
        member_service, community, members, actor, role, new_user, db):
    """Test that the given roles CAN be removed by the actor."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Delete the member again as the actor
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
    }
    member_service.delete(
        members[actor].identity,
        community._record.id,
        data
    )


def test_delete_member_type_group(
        member_service, community, owner, group, db):
    """Groups can be removed."""
    # Add a new user with role
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "reader",
    }
    member_service.add(system_identity, community._record.id, data)
    # Delete the member again
    data = {
        "members": [{"type": "group", "id": group.name}],
    }
    member_service.delete(
        owner.identity,
        community._record.id,
        data
    )


def test_delete_invalid_member(member_service, community, owner):
    """Invalid members and member types raises an error."""
    data = {"members": [{"type": "group", "id": "invalid"}]}
    pytest.raises(
        InvalidMemberError,
        member_service.delete,
        owner.identity, community._record.id, data
    )
    data = {"members": [{"type": "invalid", "id": "1"}]}
    pytest.raises(
        ValidationError,
        member_service.delete,
        owner.identity, community._record.id, data
    )


#
# Update self
#
def test_selfupdate_denied(member_service, community, any_user, new_user):
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "visible": False,
    }
    for u in [any_user, new_user]:
        pytest.raises(
            PermissionDeniedError,
            member_service.update, u.identity, community._record.id, data
        )


@pytest.mark.parametrize("actor", [
    'owner', 'manager', 'curator', 'reader',
])
def test_selfupdate_role_denied(member_service, community, members, actor):
    """Nobody can change their own role."""
    user = members[actor]
    data = {
        "members": [{"type": "user", "id": str(user.id)}],
        "role": actor,
    }
    pytest.raises(
        ValidationError,
        member_service.update, user.identity, community._record.id, data
    )


@pytest.mark.parametrize("actor", [
    'owner', 'manager', 'curator', 'reader',
])
def test_selfupdate_allow_visibility(
        member_service, community, members, actor, db):
    """All + system identity can change their own visibility to true."""
    user = members[actor]
    data = {
        "members": [{"type": "user", "id": str(user.id)}],
        "visible": True,
    }
    member_service.update(user.identity, community._record.id, data)
    member_service.update(system_identity, community._record.id, data)


#
# Update members
#
@pytest.mark.parametrize("actor,role", [
    ('owner', 'owner'),
    ('owner', 'manager'),
    ('owner', 'curator'),
    ('owner', 'reader'),
    ('manager', 'manager'),
    ('manager', 'curator'),
    ('manager', 'reader'),
])
def test_update_public_visibility_denied(
        member_service, community, members, actor, role, new_user, db):
    """Only members themselves can set public visibility."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "visible": True,
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        members[actor].identity,
        community._record.id,
        data
    )


def test_update_public_visibility_of_group_allowed(
        member_service, community, owner, group, db):
    """Group visibility can always be set to true."""
    # Add a new user with role
    data = {
        "members": [{"type": "group", "id": str(group.name)}],
        "role": "reader",
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member
    data = {
        "members": [{"type": "group", "id": str(group.name)}],
        "visible": True,
    }
    member_service.update(owner.identity, community._record.id, data)


@pytest.mark.parametrize("actor,role", [
    ('owner', 'owner'),
    ('owner', 'manager'),
    ('owner', 'curator'),
    ('owner', 'reader'),
    ('manager', 'manager'),
    ('manager', 'curator'),
    ('manager', 'reader'),
])
def test_update_hidden_visibility_allowed(
        member_service, community, members, actor, role, new_user, db):
    """Only members themselves can set public visibility."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
        "visible": True,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member with hidden visibility.
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "visible": False,
    }
    member_service.update(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize("actor,initial_role,new_role", [
    ('owner', 'owner', 'reader'),
    ('owner', 'manager', 'reader'),
    ('owner', 'curator', 'manager'),
    ('owner', 'reader', 'curator'),
    ('manager', 'manager', 'curator'),
    ('manager', 'curator', 'manager'),
    ('manager', 'reader', 'curator'),
])
def test_update_role_allowed(
        member_service, community, members, actor, initial_role, new_role,
        new_user, db):
    """Owners can change role of all, managers all but owners."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member with new role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": new_role,
    }
    member_service.update(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize("actor,initial_role,new_role", [
    ('manager', 'owner', 'reader'),
    ('curator', 'owner', 'reader'),
    ('curator', 'manager', 'curator'),
    ('curator', 'curator', 'manager'),
    ('curator', 'reader', 'owner'),
    ('reader', 'owner', 'owner'),
    ('reader', 'manager', 'manager'),
    ('reader', 'curator', 'curator'),
    ('reader', 'reader', 'manager'),
])
def test_update_role_denied(
        member_service, community, members, actor, initial_role, new_role,
        new_user, db):
    """Managers cannot change owner roles, curators/readers cannot change."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member with new role
    data["role"] = new_role
    pytest.raises(
        PermissionDeniedError,
        member_service.update,
        members[actor].identity, community._record.id, data
    )


@pytest.mark.parametrize("actor,initial_role,new_role", [
    ('owner', 'owner', 'reader'),
    ('owner', 'manager', 'reader'),
    ('owner', 'curator', 'manager'),
    ('owner', 'reader', 'curator'),
    ('manager', 'manager', 'curator'),
    ('manager', 'curator', 'manager'),
    ('manager', 'reader', 'curator'),
])
def test_update_invitation_role_allowed(
        member_service, community, members, actor, initial_role, new_role,
        new_user, db):
    """Owners can change role of all, managers all but owners."""
    # Invite a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.invite(members[actor].identity, community._record.id, data)

    # Update the invitation with new role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": new_role,
    }
    member_service.update(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize("actor,initial_role,new_role", [
    ('manager', 'owner', 'reader'),
    ('curator', 'owner', 'reader'),
    ('curator', 'manager', 'curator'),
    ('curator', 'curator', 'manager'),
    ('curator', 'reader', 'owner'),
    ('reader', 'owner', 'owner'),
    ('reader', 'manager', 'manager'),
    ('reader', 'curator', 'curator'),
    ('reader', 'reader', 'manager'),
])
def test_update_invitation_role_denied(
        member_service, community, members, actor, initial_role, new_role,
        new_user, db):
    """Managers cannot change owner roles, curators/readers cannot change."""
    # Invite a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.invite(
        members['owner'].identity, community._record.id, data)

    # Update the invitation with new role
    data["role"] = new_role
    pytest.raises(
        PermissionDeniedError,
        member_service.update,
        members[actor].identity, community._record.id, data
    )


@pytest.mark.parametrize("action", [
    'decline', 'cancel', 'expire'])
def test_update_declined_invitation(
        member_service, requests_service, community, owner, invite_user,
        invite_request_id, db, action):
    """A declined/cancelled invitation cannot be updated."""
    requests_service.execute_action(
        system_identity, invite_request_id, action).to_dict()

    # Update the invitation with new role
    data = {
        "members": [{"type": "user", "id": str(invite_user.id)}],
        "role": "reader",
    }
    assert pytest.raises(
        InvalidMemberError,
        member_service.update,
        owner.identity, community._record.id, data
    )


def test_update_role_must_have_owner(
        member_service, community, owner, group, db):
    """There must always be at least one owner."""
    # Add an owner group
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "owner",
    }
    member_service.add(system_identity, community._record.id, data)

    # Update the both owners with manager role using system_identity
    data = {
        "members": [
            {"type": "group", "id": group.name},
            {"type": "user", "id": str(owner.id)},
        ],
        "role": "manager",
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity, community._record.id, data
    )


def test_update_invalid_data(member_service, community, group):
    # No role or visibility
    data = {
        "members": [{"type": "group", "id": group.name}],
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity, community._record.id, data
    )
    # Invalid type
    data = {
        "members": [{"type": "email", "id": group.name}],
        "role": "owner"
    }
    pytest.raises(
        InvalidMemberError,
        member_service.update,
        system_identity, community._record.id, data
    )
    data = {
        "members": [{"type": "invalid", "id": group.name}],
        "role": "owner"
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity, community._record.id, data
    )
    # No members
    data = {
        "members": [],
        "role": "owner"
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity, community._record.id, data
    )
    # Missing member
    data = {
        "members": [{"type": "group", "id": "invalid"}],
        "role": "owner"
    }
    pytest.raises(
        InvalidMemberError,
        member_service.update,
        system_identity, community._record.id, data
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Communities UI views."""

from flask import Blueprint, current_app, render_template
from flask_login import current_user
from flask_menu import current_menu
from invenio_pidstore.errors import PIDDeletedError, PIDDoesNotExistError
from invenio_records_resources.services.errors import PermissionDeniedError

from .communities import communities_frontpage, communities_new, \
    communities_requests, communities_search, communities_settings, \
    communities_settings_privileges, invitations, members
from ..searchapp import search_app_context


#
# Error handlers
#
def not_found_error(error):
    """Handler for 'Not Found' errors."""
    return render_template(current_app.config['THEME_404_TEMPLATE']), 404


def record_tombstone_error(error):
    """Tombstone page."""
    return render_template("invenio_communities/tombstone.html"), 410


def record_permission_denied_error(error):
    """Handle permission denier error on record views."""
    if not current_user.is_authenticated:
        # trigger the flask-login unauthorized handler
        return current_app.login_manager.unauthorized()
    return render_template(current_app.config['THEME_403_TEMPLATE']), 403


#
# Registration
#
def create_ui_blueprint(app):
    """Register blueprint routes on app."""
    routes = app.config.get("COMMUNITIES_ROUTES")

    blueprint = Blueprint(
        "invenio_communities",
        __name__,
        template_folder="../templates",
        static_folder='../static'
    )

    # Communities URL rules
    blueprint.add_url_rule(
        routes["frontpage"],
        view_func=communities_frontpage,
    )

    blueprint.add_url_rule(
        routes["search"],
        view_func=communities_search,
    )

    blueprint.add_url_rule(
        routes["new"],
        view_func=communities_new,
    )

    # Settings tab routes
    blueprint.add_url_rule(
        routes["settings"],
        view_func=communities_settings,
    )

    blueprint.add_url_rule(
        routes["requests"],
        view_func=communities_requests,
    )

    blueprint.add_url_rule(
        routes["settings_privileges"],
        view_func=communities_settings_privileges,
    )

    blueprint.add_url_rule(
        routes['members'],
        view_func=members
    )

    blueprint.add_url_rule(
        routes['invitations'],
        view_func=invitations
    )

    @blueprint.before_app_first_request
    def register_menus():
        """Register community menu items."""
        item = current_menu.submenu('main.communities')
        item.register(
            'invenio_communities.communities_frontpage',
            'Communities',
            order=3,
        )

        item = current_menu.submenu('plus.community').register(
            'invenio_communities.communities_new',
            'New community',
            order=3,
        )

    # Register error handlers
    blueprint.register_error_handler(
        PermissionDeniedError, record_permission_denied_error)
    blueprint.register_error_handler(PIDDeletedError, record_tombstone_error)
    blueprint.register_error_handler(PIDDoesNotExistError, not_found_error)

    # Register context processor
    blueprint.app_context_processor(search_app_context)

    return blueprint



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test community service."""

from copy import deepcopy
from datetime import datetime, timedelta
import time
import pytest
from invenio_access.permissions import system_identity
from marshmallow import ValidationError
from invenio_communities.errors import CommunityFeaturedEntryDoesNotExistError
from invenio_communities.fixtures.tasks import reindex_featured_entries


@pytest.fixture()
def comm(community_service, minimal_community, location):
    """Create minimal public community."""
    c = deepcopy(minimal_community)
    c["id"] = "{id}".format(id=str(datetime.utcnow().timestamp()).replace(".","-"))
    return community_service.create(data=c, identity=system_identity)

@pytest.fixture()
def comm_restricted(community_service, minimal_community, location):
    """Create minimal restricted community."""
    c = deepcopy(minimal_community)
    c["id"] = "{id}".format(id=str(datetime.utcnow().timestamp()).replace(".","-"))
    c["access"]["visibility"] = "restricted"
    return community_service.create(data=c, identity=system_identity)


def test_search_featured(community_service, comm, db, es_clear):
    """Test that featured entries are indexed and returned correctly."""
    data = {
        "start_date": datetime.utcnow().isoformat(),
    }
    future_data = {
        "start_date": (datetime.utcnow() + timedelta(days=1)).isoformat(),
    }

    # no featured entries -> no featured communities
    featured_comms = community_service.featured_search(identity=system_identity).to_dict()
    assert len(featured_comms["hits"]["hits"]) == featured_comms["hits"]["total"] == 0

    # added one featured entry in the past. community should now be returned
    community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=data).to_dict()
    featured_comms = community_service.featured_search(identity=system_identity).to_dict()
    hits = featured_comms["hits"]["hits"]
    assert len(hits) == featured_comms["hits"]["total"] == 1
    assert hits[0]["id"] == comm.data["id"]
    # featured entries should not show up in search results
    assert "featured" not in hits[0]

    # new community with only future entry should not show up in search
    c2 = community_service.create(identity=system_identity, data={**comm.data, "id":"c2-id"})
    community_service.featured_create(identity=system_identity, community_id=c2.data["id"], data=future_data).to_dict()
    featured_comms = community_service.featured_search(identity=system_identity).to_dict()
    hits = featured_comms["hits"]["hits"]
    assert len(hits) == featured_comms["hits"]["total"] == 1
    assert hits[0]["id"] == comm.data["id"]

    # adding past featured entry to new community. first community should show up first
    data["start_date"] = (datetime.utcnow() - timedelta(days=1)).isoformat()
    community_service.featured_create(identity=system_identity, community_id=c2.data["id"], data=data).to_dict()
    featured_comms = community_service.featured_search(identity=system_identity).to_dict()
    hits = featured_comms["hits"]["hits"]
    assert len(hits) == featured_comms["hits"]["total"] == 2
    assert hits[0]["id"] == comm.data["id"]
    assert hits[1]["id"] == c2.data["id"]
    
    # adding more current past featured entry to new community. new community should show up first
    data["start_date"] = datetime.utcnow().isoformat()
    community_service.featured_create(identity=system_identity, community_id=c2.data["id"], data=data).to_dict()
    featured_comms = community_service.featured_search(identity=system_identity).to_dict()
    hits = featured_comms["hits"]["hits"]
    assert len(hits) == featured_comms["hits"]["total"] == 2
    assert hits[0]["id"] == c2.data["id"]
    assert hits[1]["id"] == comm.data["id"]


def test_reindex_featured_entries_task(community_service, comm, db, es_clear):
    """Test that reindexing task works."""
    tomorrow = {
        "start_date": (datetime.utcnow() + timedelta(days=1)).isoformat(),
    }
    c2 = community_service.create(identity=system_identity, data={**comm.data, "id":"c2-id"})
    community_service.featured_create(identity=system_identity, community_id=c2.data["id"], data=tomorrow).to_dict()

    near_future_difference=2
    near_future = {
        "start_date": (datetime.utcnow() + timedelta(seconds=near_future_difference)).isoformat(),
    }
    community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=near_future).to_dict()
    featured_comms = community_service.featured_search(identity=system_identity).to_dict()
    hits = featured_comms["hits"]["hits"]
    assert len(hits) == featured_comms["hits"]["total"] == 0

    time.sleep(near_future_difference)
    reindex_featured_entries()

    # only one community should be reindexed and returned
    featured_comms = community_service.featured_search(identity=system_identity).to_dict()
    hits = featured_comms["hits"]["hits"]
    assert len(hits) == featured_comms["hits"]["total"] == 1
    assert hits[0]["id"] == comm.data["id"]


def test_create_featured(community_service, comm, comm_restricted):
    """Test that a featured entry can be created."""
    data = {
        "start_date": datetime.utcnow().isoformat(),
    }

    f = community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=data).to_dict()
    assert f["start_date"] == data["start_date"]
    assert f["id"]

    # can not create entry for a non-public community
    with pytest.raises(ValidationError):
        community_service.featured_create(identity=system_identity, community_id=comm_restricted.data["id"], data=data)

    # can not create entry with invalid data
    with pytest.raises(ValidationError):
        community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data={})

def test_get_featured(community_service, comm):
    """Test that featured entries are indexed and returned correctly."""
    data = {
        "start_date": datetime.utcnow().isoformat(),
    }
    future_data = {
        "start_date": (datetime.utcnow() + timedelta(days=1)).isoformat(),
    }

    community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=data)
    community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=future_data)
    featured = community_service.featured_list(identity=system_identity, community_id=comm.data["id"]).to_dict()
    assert len(featured["hits"]["hits"]) == featured["hits"]["total"] == 2
    assert featured["hits"]["hits"][0]["start_date"] == data["start_date"]
    assert featured["hits"]["hits"][1]["start_date"] == future_data["start_date"]


def test_delete_featured(community_service, comm):
    """Test that featured entries are indexed and returned correctly."""
    data = {
        "start_date": datetime.utcnow().isoformat(),
    }
    future_data = {
        "start_date": (datetime.utcnow() + timedelta(days=1)).isoformat(),
    }

    past_entry = community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=data).to_dict()
    community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=future_data).to_dict()
    featured = community_service.featured_list(identity=system_identity, community_id=comm.data["id"]).to_dict()
    assert len(featured["hits"]["hits"]) == featured["hits"]["total"] == 2

    community_service.featured_delete(identity=system_identity, community_id=comm.data["id"], featured_id=past_entry["id"])
    featured = community_service.featured_list(identity=system_identity, community_id=comm.data["id"]).to_dict()
    assert len(featured["hits"]["hits"]) == featured["hits"]["total"] == 1
    assert featured["hits"]["hits"][0]["start_date"] == future_data["start_date"]

    # Error when trying to delete entry which is already deleted
    with pytest.raises(CommunityFeaturedEntryDoesNotExistError):
        community_service.featured_delete(identity=system_identity, community_id=comm.data["id"], featured_id=past_entry["id"])

    # Error when trying to delete entry which does not exist
    with pytest.raises(CommunityFeaturedEntryDoesNotExistError):
        community_service.featured_delete(identity=system_identity, community_id=comm.data["id"], featured_id=9999)

def test_update_featured(community_service, comm):
    """Test that featured entries are indexed and returned correctly."""
    data = {
        "start_date": datetime.utcnow().isoformat(),
    }
    future_data = {
        "start_date": (datetime.utcnow() + timedelta(days=1)).isoformat(),
    }

    past_entry = community_service.featured_create(identity=system_identity, community_id=comm.data["id"], data=data).to_dict()
    updated_entry = community_service.featured_update(identity=system_identity, community_id=comm.data["id"], featured_id=past_entry["id"], data=future_data).to_dict()
    assert updated_entry["id"] == past_entry["id"]
    assert updated_entry["start_date"] != past_entry["start_date"]

    featured = community_service.featured_list(identity=system_identity, community_id=comm.data["id"]).to_dict()
    assert len(featured["hits"]["hits"]) == featured["hits"]["total"] == 1
    assert featured["hits"]["hits"][0]["start_date"] == future_data["start_date"]

    # Error when trying to update entry which does not exist
    with pytest.raises(CommunityFeaturedEntryDoesNotExistError):
        community_service.featured_update(identity=system_identity, community_id=comm.data["id"], featured_id=9999, data=future_data).to_dict()
    
    # Error when trying to update entry with invalid data 
    with pytest.raises(ValidationError):
        x = community_service.featured_update(identity=system_identity, community_id=comm.data["id"], featured_id=past_entry["id"], data={}).to_dict()
        print(x)

    community_service.featured_delete(identity=system_identity, community_id=comm.data["id"], featured_id=past_entry["id"])

    # Error when trying to update entry which is already deleted
    with pytest.raises(CommunityFeaturedEntryDoesNotExistError):
        community_service.featured_update(identity=system_identity, community_id=comm.data["id"], featured_id=past_entry["id"], data=future_data).to_dict()

def test_cleanup_pre_search(db, es_clear):
    """Cleanup database and elasticsearch for following tests.
    
    Tests before do not depend on a clean db or es. Cleanup is also an
    expensive task so te tests will run faster without them.
    """
    pass



#
# This file is part of Invenio.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create featured communities table"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql
from sqlalchemy_utils import UUIDType

# revision identifiers, used by Alembic.
revision = '5b478fe7ef7f'
down_revision = 'fbe746957cfc'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        'communities_featured',
        sa.Column(
            'created',
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), 'mysql'),
            nullable=False,
        ),
        sa.Column(
            'updated',
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), 'mysql'),
            nullable=False,
        ),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('community_id', UUIDType(), nullable=False),
        sa.Column(
            'start_date',
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), 'mysql'),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ['community_id'],
            ['communities_metadata.id'],
            name=op.f(
                'fk_communities_featured_community_id_communities_metadata'
            ),
        ),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_communities_featured')),
    )


def downgrade():
    """Downgrade database."""
    op.drop_table('communities_featured')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"Members data layer API."

from invenio_accounts.models import Role
from invenio_db import db
from invenio_records.dumpers import ElasticsearchDumper
from invenio_records.dumpers.relations import RelationDumperExt
from invenio_records.systemfields import ModelField, ModelRelation, \
    RelationsField
from invenio_records_resources.records.api import Record
from invenio_records_resources.records.systemfields import IndexField
from invenio_requests.records.api import Request
from invenio_users_resources.records.api import GroupAggregate, UserAggregate
from sqlalchemy import or_

from ..errors import InvalidMemberError
from .models import ArchivedInvitationModel, MemberModel


relations_dumper = ElasticsearchDumper(extensions=[
    RelationDumperExt('relations'),
])
"""Relations dumper for members and archived invitations."""


class MemberMixin:
    """Fields defined on both member/invitation models."""

    community_id = ModelField("community_id")
    """The data-layer UUID of the community."""

    user_id = ModelField("user_id")
    """The data-layer id of the user (or None)."""

    group_id = ModelField("group_id")
    """The data-layer id of the user (or None)."""

    request_id = ModelField("request_id")
    """The data-layer id of the user (or None)."""

    role = ModelField("role")
    """The role of the entity."""

    visible = ModelField("visible")
    """Visibility of the membership."""

    active = ModelField("active")
    """Determine if it's an active membership.

    This is used for e.g. invitations where a memberships is created but not
    yet activated.
    """

    relations = RelationsField(
        user=ModelRelation(
            UserAggregate,
            'user_id',
            'user',
            attrs=[
                'email', 'username', 'profile', 'preferences', 'active',
                'confirmed',
            ],
        ),
        group=ModelRelation(
            GroupAggregate,
            'group_id',
            'group',
            attrs=['name', 'title', 'description'],
        ),
        request=ModelRelation(
            Request,
            'request_id',
            'request',
            attrs=['status', 'expires_at'],
        ),
    )

    @classmethod
    def get_memberships(cls, identity):
        """Get community memberships for a given identity."""
        # TODO: extract group/role ids from identity
        query = cls.model_cls.query_memberships(
            user_id=identity.id, group_ids=[]
        )
        return [(str(comm_id), role) for comm_id, role in query]

    @classmethod
    def get_member_by_request(cls, request_id):
        """Get a membership by request id."""
        assert request_id is not None
        obj = cls.model_cls.query.filter(
            cls.model_cls.request_id==request_id
        ).one()
        return cls(obj.data, model=obj)

    @classmethod
    def get_members(cls, community_id, members=None):
        """Get members of a community."""
        # Collect users and groups we are interested in
        user_ids = []
        group_names = []
        for m in (members or []):
            if m['type'] == 'group':
                group_names.append(m['id'])
            elif m['type'] == 'user':
                user_ids.append(m['id'])
            else:
                raise InvalidMemberError(m)

        # Query
        q = cls.model_cls.query.filter(
            cls.model_cls.community_id == community_id
        )

        # Apply user and group query if applicable
        user_q = cls.model_cls.user_id.in_(user_ids)
        groups_q = cls.model_cls.group_id.in_(
            db.session.query(Role.id).filter(Role.name.in_(group_names))
        )
        if user_ids and group_names:
            q = q.filter(or_(user_q, groups_q))
        elif user_ids:
            q = q.filter(user_q)
        elif group_names:
            q = q.filter(groups_q)

        return [cls(obj.data, model=obj) for obj in q.all()]

    @classmethod
    def has_members(cls, community_id, role=None):
        """Get members of a community."""
        return cls.model_cls.count_members(community_id, role=role)


class Member(Record, MemberMixin):
    """A member/invitation record.

    We are using a record without using the actual JSON document and
    schema validation normally used in a record. The reason for using a record
    is to facilitate the indexing which we need to have an effective search
    over the list of members.
    """

    model_cls = MemberModel

    # Needs to be here instead of on MemberMixin to overwrite Record.dumper
    dumper = relations_dumper

    # Systemfields

    metadata = None

    index = IndexField(
        "communitymembers-members-member-v1.0.0",
        search_alias="communitymembers-members"
    )
    """The ES index used."""


class ArchivedInvitation(Record, MemberMixin):
    """An archived invitation record.

    We are using a record without using the actual JSON document and
    schema validation normally used in a record. The reason for using a record
    is to facilitate the indexing which we need to have an effective search
    over the list of members.
    """

    model_cls = ArchivedInvitationModel

    # Needs to be here instead of on MemberMixin to overwrite Record.dumper
    dumper = relations_dumper

    # Systemfields

    metadata = None

    index = IndexField(
        "communitymembers-archivedinvitations-archivedinvitation-v1.0.0",
        search_alias="communitymembers"
    )
    """The ES index used."""

    @classmethod
    def create_from_member(cls, member):
        """Create an archived invitation record from a member."""
        with db.session.begin_nested():
            record = cls(
               {},
               model=cls.model_cls.from_member_model(member.model)
            )
            db.session.add(record.model)
        return record



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Member Model."""

import uuid

from invenio_accounts.models import Role, User
from invenio_db import db
from invenio_records.models import RecordMetadataBase
from invenio_requests.records.models import RequestMetadata
from sqlalchemy import CheckConstraint, Index
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy_utils.types import UUIDType

from ...communities.records.models import CommunityMetadata


class BaseMemberModel(RecordMetadataBase):
    """
    Base model for members, invitations and archived invitations.

    We restrict deletion of users/groups if they are present in the member
    table, to ensure that we have at least one owner. I.e. users must first
    be removed from memberships before they can be deleted, ensuring that
    another owner is set of a community if they are the sole owner.
    """

    id = db.Column(UUIDType, primary_key=True, default=uuid.uuid4)

    @declared_attr
    def community_id(cls):
        """Foreign key to the related community."""
        return db.Column(
            UUIDType,
            db.ForeignKey(CommunityMetadata.id, ondelete="CASCADE"),
            nullable=False
        )

    role = db.Column(db.String(50), nullable=False)

    visible = db.Column(db.Boolean(), nullable=False)

    @declared_attr
    def user_id(cls):
        """Foreign key to the related user."""
        return db.Column(
            db.Integer(),
            db.ForeignKey(User.id, ondelete="RESTRICT"),
            nullable=True,
        )

    @declared_attr
    def group_id(cls):
        """Foreign key to the related group."""
        return db.Column(
            db.Integer(),
            db.ForeignKey(Role.id, ondelete="RESTRICT"),
            nullable=True,
        )

    @declared_attr
    def request_id(cls):
        """Foreign key to the related request.

        A request can only be associated with one membership/invitation.
        """
        return db.Column(
            UUIDType,
            db.ForeignKey(RequestMetadata.id, ondelete="SET NULL"),
            nullable=True,
            unique=True,
        )

    active = db.Column(db.Boolean(), index=True, nullable=False)

    @classmethod
    def query_memberships(cls, user_id=None, group_ids=None, active=True):
        """Query for (community,role)-pairs."""
        q = db.session.query(
            cls.community_id, cls.role
        ).filter(cls.active==active)

        if user_id:
            q = q.filter(cls.user_id==user_id)
        if group_ids:
            q = q.filter(cls.group_id.in_(group_ids))

        return q.distinct()

    @classmethod
    def count_members(cls, community_id, role=None, active=True):
        """Count number of members."""
        q = cls.query.filter(
            cls.community_id==community_id,
            cls.active==active
        )
        if role is not None:
            q = q.filter(cls.role==role)
        return q.count()


class MemberModel(db.Model, BaseMemberModel):
    """Member and invitation model.

    We store members and invitations in the same table to two reasons:

    1. Reduced table size: The table is queried on login for all memberships
       of a user, and thus a smaller size is preferable.

    2. Mixing members and invitations ensures we can easily check integrity
       constraints. E.g. it's not possible to invite an existing member, and
       a person can only be invited once (database insertion will fail).
    """
    __tablename__ = "communities_members"
    __table_args__ = (
        Index('ix_community_user', 'community_id', 'user_id', unique=True),
        Index('ix_community_group', 'community_id', 'group_id', unique=True),
        # Make sure user or group is set but not both.
        CheckConstraint(
            '(user_id IS NULL AND group_id IS NOT NULL) OR '
            '(user_id IS NOT NULL AND group_id IS NULL)',
            name='user_or_group',
        ),
    )


class ArchivedInvitationModel(db.Model, BaseMemberModel):
    """Archived invitations model.

    The archived invitations model stores invitations that was rejected or
    cancelled, to support the use case of seeing if invitations was rejected
    or cancelled, and seeing past invitations.
    """

    __tablename__ = "communities_archivedinvitations"

    # We're not adding a check constraint since the row has already been
    # inserted in the member model where it was checked.

    @classmethod
    def from_member_model(cls, member_model):
        """Create an archived invitation model from a member model."""
        # Note, we keep the "active" model field, because it makes it easier to
        # handle with the search index, when we search over a combined
        # search alias for members/invitations and archived invitations.
        assert member_model.active is False
        return cls(
            id=member_model.id,
            community_id=member_model.community_id,
            user_id=member_model.user_id,
            group_id=member_model.group_id,
            request_id=member_model.request_id,
            role=member_model.role,
            visible=member_model.visible,
            active=member_model.active,
            created=member_model.created,
            updated=member_model.updated,
            version_id=member_model.version_id,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Owners classes for the access system field."""

from invenio_accounts.models import User

# TODO: Move to Invenio-Records-Resources and make reusable. Duplicated from
# Invenio-RDM-Records.
class Owner:
    """An abstraction between owner entities and specifications as dicts."""

    def __init__(self, owner):
        """Create an owner from either a dictionary or a User."""
        self._entity = None
        self.owner_type = None
        self.owner_id = None

        if isinstance(owner, dict):
            if "user" in owner:
                self.owner_type = "user"
                self.owner_id = owner["user"]

            else:
                raise ValueError("unknown owner type: {}".format(owner))

        elif isinstance(owner, User):
            self._entity = owner
            self.owner_id = owner.id
            self.owner_type = "user"

        else:
            raise TypeError("invalid owner type: {}".format(type(owner)))

    def dump(self):
        """Dump the owner to a dictionary."""
        return {self.owner_type: self.owner_id}

    def resolve(self, raise_exc=False):
        """Resolve the owner entity (e.g. User) via a database query."""
        if self._entity is None:
            if self.owner_type == "user":
                self._entity = User.query.get(self.owner_id)

            else:
                raise ValueError(
                    "unknown owner type: {}".format(self.owner_type)
                )

            if self._entity is None and raise_exc:
                raise LookupError(
                    "could not find owner: {}".format(self.dump())
                )

        return self._entity

    @property
    def user(self):
        if self.owner_type == 'user':
            return self.owner_id

    def __hash__(self):
        """Return hash(self)."""
        return hash(self.owner_type) + hash(self.owner_id)

    def __eq__(self, other):
        """Return self == other."""
        if type(self) != type(other):
            return False

        return (
            self.owner_type == other.owner_type
            and self.owner_id == other.owner_id
        )

    def __ne__(self, other):
        """Return self != other."""
        return not self == other

    def __str__(self):
        """Return str(self)."""
        return str(self.resolve())

    def __repr__(self):
        """Return repr(self)."""
        return repr(self.resolve())


class Owners(list):
    """A list of owners for a record."""

    owner_cls = Owner

    def __init__(self, owners=None, owner_cls=None):
        """Create a new list of owners."""
        self.owner_cls = owner_cls or self.owner_cls
        for owner in owners or []:
            self.add(owner)

    def add(self, owner):
        """Alias for self.append(owner)."""
        self.append(owner)

    def append(self, owner):
        """Add the specified owner to the list of owners.

        :param owner: The record's owner (either a dict, User or Owner).
        """
        if not isinstance(owner, self.owner_cls):
            owner = self.owner_cls(owner)

        if owner not in self:
            super().append(owner)

    def extend(self, owners):
        """Add all new items from the specified owners to this list."""
        for owner in owners:
            self.add(owner)

    def remove(self, owner):
        """Remove the specified owner from the list of owners.

        :param owner: The record's owner (either a dict, User or Owner).
        """
        if not isinstance(owner, self.owner_cls):
            owner = self.owner_cls(owner)

        super().remove(owner)

    def dump(self):
        """Dump the owners as a list of owner dictionaries."""
        return [owner.dump() for owner in self]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities Resource API config."""

from flask_resources import HTTPJSONException, create_error_handler
from invenio_records_resources.resources import RecordResourceConfig

community_error_handlers = RecordResourceConfig.error_handlers.copy()
community_error_handlers.update({
    FileNotFoundError: create_error_handler(
        HTTPJSONException(
            code=404,
            description="No logo exists for this community.",
        )
    ),
})


class CommunityResourceConfig(RecordResourceConfig):
    """Communities resource configuration."""

    blueprint_name = "communities"
    url_prefix = ""

    routes = {
        "communities-prefix": "/communities",
        "user-prefix": "/user/communities",
        "list": "",
        "item": "/<pid_value>",
        "featured-prefix": "/featured",
        "featured-id": "/<featured_id>",
        
    }

    error_handlers = community_error_handlers



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration."""

from datetime import timedelta

from flask_babelex import lazy_gettext as _

from invenio_communities.communities.services import facets

COMMUNITIES_ROUTES = {
    'frontpage': '/communities',
    'search': '/communities/search',
    'new': '/communities/new',
    'details': '/communities/<pid_value>',
    'settings': '/communities/<pid_value>/settings',
    'requests': '/communities/<pid_value>/requests',
    'settings_privileges': '/communities/<pid_value>/settings/privileges',
    'members': '/communities/<pid_value>/members',
    'invitations': '/communities/<pid_value>/invitations',
}

"""Communities ui endpoints."""

COMMUNITIES_ENABLED = True
"""Config to enable/disable communities blueprints."""

COMMUNITIES_ALLOW_GROUP_INVITES = True
"""Config to allow invitation of groups."""

COMMUNITIES_FACETS = {
    'type': {
        'facet': facets.type,
        'ui': {
            'field': 'type',
        }
    },
    'visibility': {
        'facet': facets.visibility,
        'ui': {
            'field': 'visibility',
        }
    }
}
"""Available facets defined for this module."""

COMMUNITIES_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_('Best match'),
        fields=['_score'],  # ES defaults to desc on `_score` field
    ),
    "newest": dict(
        title=_('Newest'),
        fields=['-created'],
    ),
    "oldest": dict(
        title=_('Oldest'),
        fields=['created'],
    ),
    "version": dict(
        title=_('Version'),
        fields=['-versions.index'],
    ),
    "updated-desc": dict(
        title=_('Recently updated'),
        fields=['-updated'],
    ),
    "updated-asc": dict(
        title=_('Least recently updated'),
        fields=['updated'],
    ),
}
"""Definitions of available record sort options."""


COMMUNITIES_ROLES = [
    dict(
        name="reader",
        title=_('Reader'),
        description=_('Can view restricted records'),
    ),
    dict(
        name="curator",
        title=_('Curator'),
        description=_('Can edit and view restricted records'),
    ),
    dict(
        name="manager",
        title=_('Manager'),
        description=_('Can manage members, curate records '
                      'and view restricted records'),
        can_manage_roles=['manager', 'curator', 'reader'],
        can_manage=True
    ),
    dict(
        name="owner",
        title=_('Owner'),
        description=_('Full administrative access to the entire community.'),
        can_manage_roles=['owner', 'manager', 'curator', 'reader'],
        is_owner=True,
        can_manage=True
    ),
]
"""Community roles."""


COMMUNITIES_REQUESTS_SEARCH = {
    'facets': ['type', 'status'],
    'sort': ['bestmatch', 'newest', 'oldest'],
}
"""Community requests search configuration (i.e list of community requests)"""

COMMUNITIES_MEMBERS_SEARCH = {
    'facets': ['role', 'visibility'],
    'sort': ['bestmatch', 'newest', 'oldest'],
}
"""Community requests search configuration (i.e list of community requests)"""

COMMUNITIES_MEMBERS_FACETS = {
    'role': {
        'facet': facets.role,
        'ui': {
            'field': 'role',
        }
    },
    'visibility': {
        'facet': facets.visible,
        'ui': {
            'field': 'visible',
        }
    }
}
"""Available facets defined for this module."""

COMMUNITIES_INVITATIONS_SEARCH = {
    'facets': ['type', 'status'],
    'sort': ['bestmatch', 'newest', 'oldest'],
}
"""Community invitations search configuration (i.e list of community invitations)"""

COMMUNITIES_INVITATIONS_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_('Best match'),
        fields=['_score'],  # ES defaults to desc on `_score` field
    ),
    "newest": dict(
        title=_('Newest'),
        fields=['-created'],
    ),
    "oldest": dict(
        title=_('Oldest'),
        fields=['created'],
    ),

}
"""Definitions of available record sort options."""


COMMUNITIES_INVITATIONS_EXPIRES_IN = timedelta(days=30)
""""Default amount of time before an invitation expires."""
